import random
import string
import uuid

from datetime import datetime
from hashlib import md5

from sqlalchemy import Boolean, Column, DateTime, Unicode, UniqueConstraint, desc, text
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import Session, relationship, synonym
from sqlalchemy.schema import ForeignKey, Table

from .meta import Base


def encrypt(val):
    return md5(val.encode("utf-8") + "v2".encode("utf-8")).hexdigest()


class LoginHistory(Base):
    __tablename__ = "login_history"
    __table_args__ = (UniqueConstraint("user_id", "device_id", "date"),)
    id = Column(
        "id", UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()"), default=uuid.uuid4
    )
    user_id = Column("user_id", UUID(as_uuid=True), ForeignKey("users.id"), nullable=False)
    device_id = Column(
        "device_id",
        UUID(as_uuid=True),
        ForeignKey("devices.id"),
        nullable=False,
    )
    date = Column("date", DateTime(timezone=True), nullable=False)

    def __init__(self, user_id=None, device_id=None, date=None, id_=None):
        self.user_id = user_id
        self.device_id = device_id
        self.date = datetime.utcnow() if date is None else date
        self.id = id_


class Device(Base):
    __tablename__ = "devices"

    id = Column(
        "id", UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()"), default=uuid.uuid4
    )
    name = Column("name", Unicode(255), unique=True, nullable=False)
    enabled = Column("enabled", Boolean, nullable=False)
    section_id = Column("section_id", UUID(as_uuid=True), ForeignKey("sections.id"), nullable=False)
    creation_date = Column("creation_date", DateTime(timezone=True), nullable=False)

    section = relationship("Section", foreign_keys=section_id)
    login_history = relationship("LoginHistory", order_by=desc(LoginHistory.date), backref="device")

    def __init__(self, name=None, enabled=None, section_id=None, creation_date=None, id_=None):
        self.name = name
        self.enabled = enabled
        self.section_id = section_id
        self.creation_date = datetime.utcnow() if creation_date is None else creation_date
        self.id = id_

    @classmethod
    def create(cls, db: Session):
        main_section = uuid.UUID("3f13f6e7-dc76-4fca-8fdb-b2bbf29b35df")
        name = "".join(random.choices(string.ascii_lowercase + string.digits, k=6))
        device = Device(name, False, main_section)
        db.add(device)
        return device


user_roles = Table(
    "user_roles",
    Base.metadata,
    Column("id", UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()"), default=uuid.uuid4),
    Column("user_id", UUID(as_uuid=True), ForeignKey("users.id"), nullable=False),
    Column("role_id", UUID(as_uuid=True), ForeignKey("roles.id"), nullable=False),
    UniqueConstraint("user_id", "role_id"),
)

role_permissions = Table(
    "role_permissions",
    Base.metadata,
    Column("id", UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()"), default=uuid.uuid4),
    Column(
        "permission_id",
        UUID(as_uuid=True),
        ForeignKey("permissions.id"),
        nullable=False,
    ),
    Column("role_id", UUID(as_uuid=True), ForeignKey("roles.id"), nullable=False),
    UniqueConstraint("permission_id", "role_id"),
)


class User(Base):
    __tablename__ = "users"

    id = Column(
        "id", UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()"), default=uuid.uuid4
    )
    name = Column("name", Unicode(255), unique=True, nullable=False)
    _password = Column("password", Unicode(60), nullable=False)
    locked_out = Column("locked_out", Boolean, nullable=False)

    roles = relationship("Role", secondary=user_roles, order_by="Role.name")
    login_history = relationship("LoginHistory", order_by=desc(LoginHistory.date), backref="user")

    def _get_password(self):
        return self._password

    def _set_password(self, password):
        self._password = encrypt(password)

    password = property(_get_password, _set_password)
    password = synonym("_password", descriptor=password)

    @property
    def __name__(self):
        return self.name

    def __init__(self, name=None, password=None, locked_out=None, id_=None):
        self.name = name
        self.password = password
        self.locked_out = locked_out
        self.id = id_

    @classmethod
    def auth(cls, name, password, db: Session):
        if password is None:
            return None
        user = db.query(User).filter(User.name.ilike(name)).first()
        if not user:
            return None
        if user.password != encrypt(password) or user.locked_out:
            return None
        else:
            return user


class Role(Base):
    __tablename__ = "roles"

    id = Column(
        "id", UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()"), default=uuid.uuid4
    )
    name = Column("name", Unicode(255), unique=True)

    def __init__(self, name=None, id_=None):
        self.name = name
        self.id = id_


class Permission(Base):
    __tablename__ = "permissions"

    id = Column(
        "id", UUID(as_uuid=True), primary_key=True, server_default=text("gen_random_uuid()"), default=uuid.uuid4
    )
    name = Column("name", Unicode(255), unique=True)

    roles = relationship("Role", secondary=role_permissions, backref="permissions")

    def __init__(self, name=None, id_=None):
        self.name = name
        self.id = id_
