import { HttpClient, HttpHeaders, HttpParams } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { Observable } from 'rxjs/internal/Observable';
import { catchError } from 'rxjs/operators';

import { ErrorLoggerService } from '../core/error-logger.service';
import { Product } from '../core/product';

const httpOptions = {
  headers: new HttpHeaders({ 'Content-Type': 'application/json' }),
};

const url = '/api/products';
const serviceName = 'ProductService';

@Injectable({ providedIn: 'root' })
export class ProductService {
  constructor(private http: HttpClient, private log: ErrorLoggerService) {}

  get(id: string | null): Observable<Product> {
    const getUrl: string = id === null ? `${url}` : `${url}/${id}`;
    return <Observable<Product>>(
      this.http
        .get<Product>(getUrl)
        .pipe(catchError(this.log.handleError(serviceName, `get id=${id}`)))
    );
  }

  list(): Observable<Product[]> {
    return <Observable<Product[]>>(
      this.http
        .get<Product[]>(`${url}/list`)
        .pipe(catchError(this.log.handleError(serviceName, 'list')))
    );
  }

  listIsActiveOfCategory(id: string): Observable<Product[]> {
    const options = { params: new HttpParams().set('mc', id) };
    return <Observable<Product[]>>(
      this.http
        .get<Product[]>(`${url}/query`, options)
        .pipe(catchError(this.log.handleError(serviceName, 'listIsActiveOfCategory')))
    );
  }

  save(product: Product): Observable<Product> {
    return <Observable<Product>>(
      this.http
        .post<Product>(`${url}`, product, httpOptions)
        .pipe(catchError(this.log.handleError(serviceName, 'save')))
    );
  }

  update(product: Product): Observable<Product> {
    return <Observable<Product>>(
      this.http
        .put<Product>(`${url}/${product.id}`, product, httpOptions)
        .pipe(catchError(this.log.handleError(serviceName, 'update')))
    );
  }

  updateSortOrder(list: Product[]): Observable<Product[]> {
    return <Observable<Product[]>>(
      this.http
        .post<Product[]>(`${url}/list`, list, httpOptions)
        .pipe(catchError(this.log.handleError(serviceName, 'updateSortOrder')))
    );
  }

  saveOrUpdate(product: Product): Observable<Product> {
    if (!product.id) {
      return this.save(product);
    }
    return this.update(product);
  }

  delete(id: string): Observable<Product> {
    return <Observable<Product>>(
      this.http
        .delete<Product>(`${url}/${id}`, httpOptions)
        .pipe(catchError(this.log.handleError(serviceName, 'delete')))
    );
  }

  balance(id: string, date: string): Observable<number> {
    const options = { params: new HttpParams().set('d', date) };
    return <Observable<number>>(
      this.http
        .get<number>(`${url}/balance`, options)
        .pipe(catchError(this.log.handleError(serviceName, 'balance')))
    );
  }
}
